/*	Cory McWilliams
		mcwilc@rpi.edu
		10/2/2003 */

#if defined (__APPLE__)
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#include <cmath>
#include <iostream>
#include "objects.h"
#include "vertex.h"

#define ABS(a) ((a)>0?(a):(-a))

// Global state variables
int viewMode = 0;
float trainTime;
float trainSep = 0.02f;
float rot[3];
bool buttons[3];
int ww;
int wh;
int lastx;
int lasty;
float position[3] = {0.0f, 0.75f, 5.0f};
bool keys[256];
float gate[2];
float wobble;
float carPos[3] = {3.0f, 0.0f, 3.0f};
float carRot = 20.0f;
float carSpeed;

// Data to help define the scene
const int trackCount = 10;
const int buildingCount = 6;
const int roadCount = 6;
const int treeCount = 13;
float tracks[trackCount][3] = {
	{14.41f, 0.0f, 0.0f},
	{34.41f, 0.0f, 0.0f},
	{54.41f, 0.0f, 0.0f},
	{68.82f, 0.0f, 14.41f},
	{68.82f, 0.0f, 34.41f},
	{54.41f, 0.0f, 48.82f},
	{34.41f, 0.0f, 48.82f},
	{14.41f, 0.0f, 48.82f},
	{0.0f, 0.0f, 34.41f},
	{0.0f, 0.0f, 14.41f},
};
float roads[roadCount][3] = {
	{24.41f, 0.0f, 19.41f},
	{73.28f, 0.0f, 38.28f},
	{59.41f, 0.0f, 75.41f},
	{44.41f, 0.0f, 65.41f},
	{23.41f, 0.0f, 75.41f},
	{22.41f, 0.0f, 38.28f},
};
float trees[treeCount][2] = {
	{14.0703, 7.97066},
	{30.0353, 6.26715},
	{40.0877, 16.6484},
	{52.6321, 32.2672},
	{63.2237, 52.7061},
	{66.4488, 75.5082},
	{46.7225, 72.6206},
	{38.6267, 56.5957},
	{28.9317, 56.1292},
	{15.2921, 55.6688},
	{14.5458, 41.1735},
	{9.38945, 26.7794},
	{0.0f, 0.0f}
};
float buildings[buildingCount][5] = {
	{15.0f, 15.0f, 5.0f, 1.0f, 1.0f},
	{15.0f, 35.0f, 2.0f, 1.3f, 2.0f},
	{35.0f, 16.0f, 3.0f, 3.0f, 3.0f},
	{33.0f, 36.0f, 9.0f, 3.0f, 4.0f},
	{56.0f, 65.0f, 7.0f, 5.0f, 5.0f},
	{55.0f, 16.0f, 10.0f, 1.0f, 6.0f}
};

// Handle one particular potential collision.
void collide(vertex &cbig, vertex &csmall, float radius) {
	vertex diff = csmall - cbig;
	float mag = abs(diff);
	if (mag < radius) {
		// Just move the smaller object away if it's offending
		csmall += normalize(diff) * (radius - mag);
	}
}

/* position[3], the train's position, and carPos[3] can not intersect
	 position[3] and carPos[3] can't intersect trees or buildings
	 carPos[3] can't get near the gates when they're down */
void collisions() {
	// Get the positions of the main objects
	vertex hp = {position[0], 0.0f, position[2]};
	vertex cp = {carPos[0], 0.0f, carPos[2]};

	collide(cp, hp, 1.5f); // person can't hit the car
	for (int i = 0; i < 5; i++) { // Look at each train car individuall
		vertex tcp;
		catmullRomSpline((float *)tracks, trackCount, trainTime - trainSep * i, (float *)&tcp);
		collide(tcp, hp, 2.0f); // person can't hit the train
		collide(tcp, cp, 2.5f);	// car can't hit the train
	}

	// Hit buildings
	for (int i = 0; i < buildingCount; i++) {
		vertex b = {buildings[i][0], 0.0f, buildings[i][1]};
		float size = buildings[i][2] * 0.5f * 1.414f;
		collide(b, hp, size + 0.5f);
		collide(b, cp, size + 1.0f);
	}

	// The little train gates...(they don't apply to pedestrians)
	vertex g1 = {25.14f, 0.0f, 51.28f};
	vertex g2 = {20.14f, 0.0f, 48.28f};
	vertex g3 = {71.0f, 0.0f, 32.5f};
	vertex g4 = {67.0f, 0.0f, 35.0f};
	collide(g1, cp, 1.5f); // collide with the actual gates
	collide(g2, cp, 1.5f);
	collide(g3, cp, 1.5f);
	collide(g4, cp, 1.5f);
	vertex i1 = (g1 + g2) * 0.5f;
	vertex i2 = (g3 + g4) * 0.5f;
	if (ABS(gate[0]) < 45.0f) collide(i1, cp, 5.0f); // don't be in a closed intersection
	if (ABS(gate[1]) < 45.0f) collide(i2, cp, 5.0f);

	// hit trees
	for (int i = 0; i < treeCount; i++) {
		vertex t = {trees[i][0], 0.0f, trees[i][1]};
		collide(t, cp, 1.25f);
		collide(t, hp, 0.5f);
	}

	// Actually move the objects
	position[0] = hp.x;
	position[2] = hp.z;
	carPos[0] = cp.x;
	carPos[2] = cp.z;
}

// Some simple initialization
void init() {
	glClearColor(0.0f, 0.0f, 1.0f, 1.0f);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE); /* I do some scaling and other transformations which require that
		I make OpenGL normalize my vectors, or else the lighting gets pretty bad */

	// Lighting
	float white[4] = {1.0f, 1.0f, 1.0f, 0.0f};
	float ambient[4] = {0.1f, 0.1f, 0.1f, 0.0f};
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, white);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
	glMaterialfv(GL_FRONT, GL_SPECULAR, white);
}

// Beautiful display routine
void display() {
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glPushMatrix();

	switch (viewMode) {
	case 0: { // Walking mode...a first person camera that kind of wavers as it moves
		float walk[3] = {0.0f, 0.0f, 0.0f};
		float theta = float(rot[1] * M_PI / 180.0f);
		float waver = 0.1f * cos(wobble * M_PI / 180.0f);
		walk[0] = waver * cos(theta);
		walk[1] = 0.05f * (1.0f + sin(2 * wobble * M_PI / 180.0f));
		walk[2] = waver * sin(theta);
		glLoadIdentity();
		glRotatef(rot[0], 1.0f, 0.0f, 0.0f);
		glRotatef(rot[1], 0.0f, 1.0f, 0.0f);
		glRotatef(rot[2], 0.0f, 0.0f, 1.0f);
		glTranslatef(-position[0] - walk[0], -position[1] - walk[1], -position[2] - walk[2]);
		break; }
	case 1: { // Train mode...ride the train engine around the track
		glLoadIdentity();
		float trainPos[3];
		float ahead[3];
		catmullRomSpline((float *)tracks, trackCount, trainTime, (float *)trainPos);
		catmullRomSpline((float *)tracks, trackCount, trainTime + 0.02f, (float *)ahead);
		gluLookAt(trainPos[0], trainPos[1] + 3.0f, trainPos[2],
			ahead[0], ahead[1] + 2.0f, ahead[2],
			0.0f, 1.0f, 0.0f);
		break; }
	case 2: // Car mode...drive the car around
		glLoadIdentity();
		glRotatef(carRot, 0.0f, 1.0f, 0.0f);
		glTranslatef(-carPos[0], -carPos[1] - 1.0f, -carPos[2]);
		break;
	}

	// Light position
	float pos[4] = {20.0f, 30.0f, 20.0f, 0.0f};
	glLightfv(GL_LIGHT0, GL_POSITION, pos);

	// Draw the ground first
	float grass[4] = {0.0f, 0.4f, 0.0f, 0.0f};
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, grass);
	glNormal3f(0.0f, 1.0f, 0.0f);
	glBegin(GL_QUADS);
	glVertex3f(-40.0f, 0.0f, -40.0f);
	glVertex3f(130.0f, 0.0f, -40.0f);
	glVertex3f(130.0f, 0.0f, 100.0f);
	glVertex3f(-40.0f, 0.0f, 100.0f);
	glEnd();

	/* The train tracks and road are drawn in the same plane as the grass, so it is
		necessary to disable depth testing to be able to draw them correctly. */
	glDisable(GL_DEPTH_TEST);
	glLineWidth(2.0f);
	road(*roads, roadCount); // road
	trainTracks(*tracks, trackCount); // tracks
	glEnable(GL_DEPTH_TEST);

	// Draw some trees spotting the landscape
	for (int i = 0; i < treeCount; i++)
		tree(trees[i][0], trees[i][1]);

	// Draw the car in its place
	glPushMatrix();
	glTranslatef(carPos[0], carPos[1], carPos[2]);
	glRotatef(-carRot + 90.0f, 0.0f, 1.0f, 0.0f);
	car();
	glPopMatrix();

	// Draw lots of buildings
	for (int i = 0; i < buildingCount; i++)
		building(buildings[i][0], buildings[i][1], buildings[i][2],
			buildings[i][3], buildings[i][4]);

	// Draw the train cars
	for (int car = 0; car < 5; car++) {
		glPushMatrix();
		float tpos[3];
		float tgrad[3];
		float trainCarTime = trainTime - car * trainSep;
		if (trainCarTime < 0.0f) trainCarTime += 1.0f;

		/* I'm using these spiffy Catmull-Rom splines I discovered a while ago
			to achieve a lot of nice effects. */
		catmullRomSpline((float *)tracks, trackCount, trainCarTime, (float *)tpos);
		catmullRomGradient((float *)tracks, trackCount, trainCarTime, (float *)tgrad);

		glTranslatef(tpos[0], tpos[1], tpos[2]);
		vertex tvec = {-1.0f, 0.0f, 0.0f};
		vertex ttan = {-tgrad[0], -tgrad[1], -tgrad[2]};

		// Find angle for rotation from the slope...
		float theta = acos((ttan.x - tvec.x) / abs(tvec - ttan)) * 180.0f / M_PI;
		if (ttan.z > 0.0f)
			theta *= -1.0f;

		glRotatef(theta, 0.0f, 1.0f, 0.0f);

		// Engine, box car, box car, flat bed, box car
		if (car == 0)
			trainEngine();
		else if (car == 3)
			trainFlatBed();
		else
			train();
		glPopMatrix();
	}

	// Cross gates
	glPushMatrix();
	glTranslatef(25.14f, 0.0f, 51.28f);
	crossGate(ABS(gate[0]));
	glPopMatrix();
	glPushMatrix();
	glTranslatef(20.14f, 0.0f, 48.28f);
	glRotatef(180.0f, 0.0f, 1.0f, 0.0f);
	crossGate(ABS(gate[0]));
	glPopMatrix();

	// Another pair of cross gates
	glPushMatrix();
	glTranslatef(71.0f, 0.0f, 32.5f);
	glRotatef(60.0f, 0.0f, 1.0f, 0.0f);
	crossGate(ABS(gate[1]));
	glPopMatrix();
	glPushMatrix();
	glTranslatef(67.0f, 0.0f, 35.0f);
	glRotatef(250.0f, 0.0f, 1.0f, 0.0f);
	crossGate(ABS(gate[1]));
	glPopMatrix();

	glPopMatrix();
	glutSwapBuffers();
}

// Simple perspective
void reshape(int width, int height) {
	ww = width;
	wh = height;
	if (!height) height=1; // prevent division by zero
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0f, float(width)/float(height), 0.01f, 100.0f);
	glMatrixMode(GL_MODELVIEW);
}

// When you're walking, you can drag the view around to turn
void motion(int x, int y) {
	if (buttons[0]) {
		rot[0] -= (90.0f * (y - lasty)) / wh;
		rot[1] -= (90.0f * (x - lastx)) / ww;
		glutPostRedisplay();
	}
	lastx = x; lasty = y;
}

// Maintain a table of the state of each button
void mouse(int button, int state, int x, int y) {
	buttons[button] = (state == GLUT_DOWN);
	lastx = x; lasty = y;
}

// Keypress events are used to maintain the state of each key and to get into and out of vehicles
void keyboard(unsigned char key, int x, int y) {
	keys[key] = true;
	switch (key) {
	case '\r':
	case '\n': { // Enter lets you get into and out of the car and train
		vertex humanPos = {position[0], position[1], position[2]};
		vertex trainPos;
		vertex cPos = {carPos[0], carPos[1], carPos[2]};
		catmullRomSpline((float *)tracks, trackCount, trainTime, (float *)&trainPos);
		if (viewMode == 0 && abs(humanPos - trainPos) < 4.0f) {
			viewMode = 1;
		} else if (viewMode == 0 && abs(humanPos - cPos) < 3.0f) {
			viewMode = 2;
		} else if (viewMode == 1) {
			vertex grad;
			catmullRomGradient((float *)tracks, trackCount, trainTime, (float *)&grad);
			vertex up = {0.0f, 1.0f,0.0f};
			vertex aside = normalize(crossProduct(up, grad));
			position[0] = trainPos.x + aside.x;
			position[2] = trainPos.z + aside.z;
			viewMode = 0;
		} else if (viewMode == 2) {
			position[0] = cPos.x + 1.0f;
			position[2] = cPos.z + 1.0f;
			viewMode = 0;
			carSpeed = 0.0f;
		}
		break; }
	}
}

// Update key table
void keyboardUp(unsigned char key, int x, int y) {
	keys[key] = false;
}

// The "real" work
void timer(int t) {
	// Move the person...but it's not that easy, because it is necessary to take angle into account
	float step = 0.100f;
	float theta = float(rot[1] * M_PI / 180.0f);
	float zstep = cos(theta) * step;
	float xstep = sin(theta) * step;
	if (keys['w']) { position[0] += xstep; position[2] -= zstep; }
	if (keys['s']) { position[0] -= xstep; position[2] += zstep; }
	if (keys['d']) { position[0] += zstep; position[2] += xstep; }
	if (keys['a']) { position[0] -= zstep; position[2] -= xstep; }
	if (keys['q']) rot[1] -= 45.0f * step;
	if (keys['e']) rot[1] += 45.0f * step;

	// Move the gates up and down when the train is near
	if (trainTime > 0.5f && trainTime < 0.7f) {
		if (gate[0] > 0.0f) gate[0] -= 1.0f;
		if (gate[0] < 0.0f) gate[0] = 0.0f;
	} else {
		if (gate[0] < 90.0f) gate[0] += 1.0f;
		if (gate[0] > 90.0f) gate[0] = 90.0f;
	}
	if (trainTime > 0.2f && trainTime < 0.4f) {
		if (gate[1] > 0.0f) gate[1] -= 1.0f;
		if (gate[1] < 0.0f) gate[1] = 0.0f;
	} else {
		if (gate[1] < 90.0f) gate[1] += 1.0f;
		if (gate[1] > 90.0f) gate[1] = 90.0f;
	}

	// If the camera is walking, make it waver around a little
	if (keys['a'] || keys['w'] || keys['s'] || keys['d']) {
		wobble += 10.0f;
		if (wobble > 360.0f) wobble = 0.0f;
	}

	// If the car is driving, update it's speed and direction, with some reasonable extremes
	if (viewMode == 2) {
		if (keys['w']) // forward
			carSpeed += 0.1f;
		if (keys['s']) // reverse
			carSpeed -= 0.1f;

		// Turn, but invert the steering properly
		if ((keys['d'] && carSpeed > 0) || (keys['a'] && carSpeed < 0)) { // right
			carRot += 1.0f;
			if (carRot > 360.0f) carRot = 0.0f;
		}
		if ((keys['a'] && carSpeed > 0) || (keys['d'] && carSpeed < 0)) { // left
			carRot -= 1.0f;
			if (carRot < 0.0f) carRot = 360.0f;
		}
		if (keys[' '] && carSpeed) { // brake
			carSpeed *= 0.9f;
		}
		carSpeed *= 0.99f; // Friction
		if (carSpeed > 3.0f) carSpeed = 3.0f; // Speed limit
		if (carSpeed < -2.0f) carSpeed = -2.0f; // Reverse speed limit
		// Slowing down could go on for a long time...put it to a stop if it gets too slow
		if (ABS(carSpeed) < 0.001f) carSpeed = 0.0f;
	}

	// Update the car's position
	if (viewMode == 2 && carSpeed) {
		const float step = 0.1f * carSpeed;
		float theta = float(carRot * M_PI / 180.0f);
		float zstep = cos(theta) * step;
		float xstep = sin(theta) * step;
		carPos[0] += xstep;
		carPos[2] -= zstep;
	}

	// If riding the train, let the user control it...a little
	if (viewMode != 1 || keys['w']) {
		trainTime += 0.0003f;
		if (trainTime >= 1.0f) trainTime = 0.0f;
	}
	if (viewMode == 1 && keys['s']) {
		trainTime -= 0.0003f;
		if (trainTime < 0.0f) trainTime = 1.0f;
	}

	// Handle all of the possible collisions
	collisions();

	// Set another timer and redraw
	glutPostRedisplay();
	glutTimerFunc(10, timer, 0);
}

void joystick(unsigned int buttons, int x, int y, int z) {
	static unsigned int lastButtons = buttons;
	static int lastX = x;
	static int lastY = y;

	if (x != lastX) {
		if (x > 0) {
			keys['d'] = true;
		} else if (x < 0) {
			keys['a'] = true;
		} else {
			keys['d'] = keys['a'] = false;
		}
		lastX = x;
	}
	if (y != lastY) {
		if (y > 0) {
			keys['s'] = true;
		} else if (y < 0) {
			keys['w'] = true;
		} else {
			keys['s'] = keys['w'] = false;
		}
		lastY = y;
	}

	if (lastButtons != buttons) {
		keys['\n'] = (buttons & (1 << 0)) != 0;
		keys['q'] = (buttons & (1 << 1)) != 0;
		keys['e'] = (buttons & (1 << 2)) != 0;
		keys[' '] = (buttons & (1 << 3)) != 0;
		if (keys['\n']) {
			keyboard('\n', 0, 0);
		}
		lastButtons = buttons;
	}
}

// Program entry point
int main(int argc, char *argv[]) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB|GLUT_DOUBLE|GLUT_DEPTH);
	glutInitWindowSize(640, 480);
	glutCreateWindow(argv[0]);
	init();
	glutIgnoreKeyRepeat(true);
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutTimerFunc(0, timer, 0);
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyboardUp);
	glutJoystickFunc(joystick, 10);
	glutFullScreen();
	glutSetCursor(GLUT_CURSOR_NONE);
	glutMainLoop();
	return 0;
}
