/*	Cory McWilliams
		mcwilc@rpi.edu
		10/2/2003 */

#include "objects.h"
#if defined (__APPLE__)
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#include <math.h>
#include <iostream>
#include "vertex.h"

GLUquadric *gluq;

// Some sensible constants
const float trainShine = 32.0f;
const float dullShine = 1.0f;
const float black[4] = {0.0f, 0.0f, 0.0f, 0.0f};
const float white[4] = {1.0f, 1.0f, 1.0f, 0.0f};

// For convenience, a filled cylinder with only two parameters
void cylinder(float radius, float depth) {
	if (!gluq) gluq = gluNewQuadric();
	if (!gluq) return;

	glPushMatrix();
	glTranslatef(0.0f, 0.0f, -depth / 2);
	gluCylinder(gluq, radius, radius, depth, 12, 1);
	glTranslatef(0.0f, 0.0f, depth);
	gluDisk(gluq, 0.0f, radius, 12, 1);
	glTranslatef(0.0f, 0.0f, -depth);
	glRotatef(180.0f, 1.0f, 0.0f, 0.0f);
	gluDisk(gluq, 0.0f, radius, 12, 1);
	glPopMatrix();
}

// Draw a shiny metal box car
void train() {
	float red[4] = {1.0f, 0.0f, 0.0f, 0.0f};
	float black[4] = {0.2f, 0.2f, 0.2f, 0.0f};
	
	glMaterialfv(GL_FRONT, GL_SHININESS, &trainShine);
	glMaterialfv(GL_FRONT, GL_SPECULAR, white);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, red);

	glPushMatrix();
	glScalef(3.0f, 1.0f, 1.0f);
	glTranslatef(0.0f, 1.0f, 0.0f);
	glutSolidCube(1.0f);
	glPopMatrix();

	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, black);
	glPushMatrix();
	glTranslatef(1.25f, 0.25f, 0.45f);
	cylinder(0.25f, 0.1f);
	glPopMatrix();
	
	glPushMatrix();
	glTranslatef(-1.25f, 0.25, 0.45f);
	cylinder(0.25f, 0.1f);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(1.25f, 0.25f, -0.45f);
	cylinder(0.25f, 0.1f);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(-1.25f, 0.25f, -0.45f);
	cylinder(0.25f, 0.1f);
	glPopMatrix();
}

// Draw a shiny metal Ford Mustang Convertible...er, a box with four wheels
void car() {
	float blue[4] = {0.0f, 0.0f, 1.0f, 0.0f};
	float black[4] = {0.2f, 0.2f, 0.2f, 0.0f};
	
	glMaterialfv(GL_FRONT, GL_SHININESS, &trainShine);
	glMaterialfv(GL_FRONT, GL_SPECULAR, white);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, blue);

	glPushMatrix();
	glScalef(1.75f, 0.5f, 0.6f);
	glTranslatef(0.0f, 1.0f, 0.0f);
	glutSolidCube(1.0f);
	glPopMatrix();

	glMaterialfv(GL_FRONT, GL_SPECULAR, black);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, black);
	glPushMatrix();
	glTranslatef(0.75f, 0.25f, 0.35f);
	cylinder(0.25f, 0.1f);
	glPopMatrix();
	
	glPushMatrix();
	glTranslatef(-0.75f, 0.25, 0.35f);
	cylinder(0.25f, 0.1f);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(0.75f, 0.25f, -0.35f);
	cylinder(0.25f, 0.1f);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(-0.75f, 0.25f, -0.35f);
	cylinder(0.25f, 0.1f);
	glPopMatrix();
}

// A flatbed train car.
void trainFlatBed() {
	float red[4] = {1.0f, 0.0f, 0.0f, 0.0f};
	float black[4] = {0.2f, 0.2f, 0.2f, 0.0f};
	glMaterialfv(GL_FRONT, GL_SHININESS, &trainShine);
	glMaterialfv(GL_FRONT, GL_SPECULAR, white);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, red);

	glPushMatrix();
	glTranslatef(0.0f, 0.625f, 0.0f);
	glScalef(3.0f, 0.25f, 1.0f);
	glutSolidCube(1.0f);
	glPopMatrix();

	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, black);
	glPushMatrix();
	glTranslatef(1.25f, 0.25f, 0.45f);
	glColor3f(0.0f, 1.0f, 0.0f);
	cylinder(0.25f, 0.1f);
	glPopMatrix();
	
	glPushMatrix();
	glTranslatef(-1.25f, 0.25, 0.45f);
	glColor3f(0.0f, 1.0f, 0.0f);
	cylinder(0.25f, 0.1f);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(1.25f, 0.25f, -0.45f);
	glColor3f(0.0f, 1.0f, 0.0f);
	cylinder(0.25f, 0.1f);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(-1.25f, 0.25f, -0.45f);
	glColor3f(0.0f, 1.0f, 0.0f);
	cylinder(0.25f, 0.1f);
	glPopMatrix();
}

// A train engine.  My only regret is that it doesn't have smoke
void trainEngine() {
	trainFlatBed();
	float red[4] = {1.0f, 0.0f, 0.0f, 0.0f};
	glMaterialfv(GL_FRONT, GL_SHININESS, &trainShine);
	glMaterialfv(GL_FRONT, GL_SPECULAR, white);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, red);
	glPushMatrix();
	glTranslatef(1.0f, 1.25f, 0.0f);
	glutSolidCube(1.0f);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(-0.5f, 1.0f, 0.0f);
	glRotatef(90.0f, 0.0f, 1.0f, 0.0f);
	cylinder(0.35f, 2.0f);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(-1.2f, 1.5f, 0.0f);
	glRotatef(90.0f, 1.0f, 0.0f, 0.0f);
	cylinder(0.2f, 1.0f);
	glPopMatrix();
}

// A gate to block the road when a train is near
void crossGate(float angle) {
	float metal[4] = {0.5f, 0.5f, 0.5f, 0.0f};
	float red[4] = {1.0f, 0.0f, 0.0f, 0.0f};
	float white[4] = {1.0f, 1.0f, 1.0f, 0.0f};
	angle = -angle;
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, metal);
	glMaterialfv(GL_FRONT, GL_SHININESS, &dullShine);
	glMaterialfv(GL_FRONT, GL_SPECULAR, black);

	glPushMatrix();
	glScalef(1.0f, 0.1f, 1.0f);
	glTranslatef(0.0f, 0.5f, 0.0f);
	glutSolidCube(0.5f);
	glPopMatrix();
	glPushMatrix();
	glScalef(0.1f, 1.5f, 0.1f);
	glTranslatef(0.0f, 0.5f, 0.0f);
	glutSolidCube(1.0f);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(-0.2f, 1.0f, 0.0f);
	glPushMatrix();
	glScalef(1.0f, 0.5f, 0.1f);
	glutSolidCube(0.4f);
	glPopMatrix();
	glTranslatef(-0.11f, 0.0f, 0.0f);
	glRotatef(angle, 0.0f, 0.0f, 1.0f);
	glScalef(1.5f, 1.0f, 0.1f);
	for (int i = 0; i < 10; i++) { // Lots of alternating red and white cubes to make a gate
		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, (i%2)?red:white);
		glutSolidCube(0.2f);
		glTranslatef(-0.2f, 0.0f, 0.0f);
	}
	glPopMatrix();
}

// Guided by control points and a Catmull-Rom spline, draw two rails and lots of ties
void trainTracks(float *tracks, int trackCount) {
	const float step = 0.005f;
	glDisable(GL_LIGHTING);

	glColor3f(0.0f, 0.0f, 0.0f);
	for (float t = 0; t <= 1.0f; t += step) {
		/* Follow the spline, finding the slope of the spline and then finding
			a point on each side between which to place a railroad tie. */
		float v[3];
		float d[3];
		vertex up = {0.0f, 1.0f, 0.0f};
		catmullRomSpline(tracks, trackCount, t, v);
		catmullRomGradient(tracks, trackCount, t, d);
		vertex p = {v[0], v[1], v[2]};
		vertex g = {d[0], d[1], d[2]};
		g = normalize(crossProduct(up, g));
		vertex f = p - 0.5f * g;
		g = p + 0.5f * g;
		glBegin(GL_LINES);
		glVertex3fv((float *)&f);
		glVertex3fv((float *)&g);
		glEnd();
	}

	// Go down one side of the spline and place a rail
	glBegin(GL_LINE_LOOP);
	for (float t = 0; t <= 1.0f; t += step) {
		float v[3];
		float d[3];
		vertex up = {0.0f, 1.0f, 0.0f};
		catmullRomSpline(tracks, trackCount, t, v);
		catmullRomGradient(tracks, trackCount, t, d);
		vertex p = {v[0], v[1], v[2]};
		vertex g = {d[0], d[1], d[2]};
		g = normalize(crossProduct(up, g));
		vertex f = p - 0.5f * g;
		g = p + 0.5f * g;
		glVertex3fv((float *)&f);
	}
	glEnd();
	// Go down the other side of the spline and place a rail
	glBegin(GL_LINE_LOOP);
	for (float t = 0; t <= 1.0f; t += step) {
		float v[3];
		float d[3];
		vertex up = {0.0f, 1.0f, 0.0f};
		catmullRomSpline(tracks, trackCount, t, v);
		catmullRomGradient(tracks, trackCount, t, d);
		vertex p = {v[0], v[1], v[2]};
		vertex g = {d[0], d[1], d[2]};
		g = normalize(crossProduct(up, g));
		g = p + 0.5f * g;
		glVertex3fv((float *)&g);
	}
	glEnd();
	glEnable(GL_LIGHTING);
}

// A box, kind of dull, and of a weird color
void building(float x, float z, float s, float h, float d) {
	int c = int(d);
	float bright = 0.75f;
	float dark = 0.25f;
	float color[3] = {(c&1)?bright:dark,
		(c&2)?bright:dark, (c&4)?bright:dark};
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color);
	glMaterialfv(GL_FRONT, GL_SHININESS, &dullShine);
	glMaterialfv(GL_FRONT, GL_SPECULAR, black);
	glPushMatrix();
	glTranslatef(x, s / 2, z);
	glutSolidCube(s);
	glPopMatrix();
}

// A pine tree
void tree(float x, float z) {
	if (!gluq) gluq = gluNewQuadric();
	if (!gluq) return;

	float wood[4] = {0.6f, 0.3f, 0.0f, 0.0f};
	float leaves[4] = {0.0f, 0.5f, 0.0f, 0.0f};
	const int parts = 5;

	glMaterialfv(GL_FRONT, GL_SHININESS, &dullShine);
	glMaterialfv(GL_FRONT, GL_SPECULAR, black);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, wood);
	glPushMatrix();
	glTranslatef(x, 0.0f, z);
	glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);
	gluCylinder(gluq, 0.1f, 0.1f, 0.5f, parts, 1); // trunk
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, leaves);
	glTranslatef(0.0f, 0.0f, 0.5f);
	
	gluCylinder(gluq, 0.5f, 0.3f, 0.5f, parts, 1); // tier 1
	gluDisk(gluq, 0.0f, 0.5f, parts, 1);
	glTranslatef(0.0f, 0.0f, 0.5f);
	
	gluCylinder(gluq, 0.4f, 0.2f, 0.5f, parts, 1); // tier 2
	gluDisk(gluq, 0.0f, 0.4f, parts, 1);
	glTranslatef(0.0f, 0.0f, 0.5f);
	
	gluCylinder(gluq, 0.3f, 0.0f, 0.5f, parts, 1); // tier 3
	gluDisk(gluq, 0.0f, 0.3f, parts, 1);
	glPopMatrix();
}

// The road, a gray path
void road(float *points, int pointCount) {
	const float step = 0.01f;
	float gravel[4] = {0.2f, 0.2f, 0.2f, 0.0f};

	glMaterialfv(GL_FRONT, GL_SHININESS, &dullShine);
	glMaterialfv(GL_FRONT, GL_SPECULAR, black);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gravel);
	glBegin(GL_QUAD_STRIP);
	for (float t = 0; t <= 1.0f; t += step) {
		// Go an arbitrary number of steps down the path, drawing road
		float v[3];
		float d[3];
		vertex up = {0.0f, 1.0f, 0.0f};
		catmullRomSpline(points, pointCount, t, v);
		catmullRomGradient(points, pointCount, t, d);
		vertex p = {v[0], v[1], v[2]};
		vertex g = {d[0], d[1], d[2]};
		g = normalize(crossProduct(up, g));
		vertex f = p - g;
		g = p + g;
		glVertex3fv((float *)&f);
		glVertex3fv((float *)&g);
	}
	glEnd();
}
